#include "Arduino.h"
#include "ESPNow.hpp"
#include "NoDataAvailableException.hpp"
#include "SentecRainGaugeSensor.h"
#include "rs485.hpp"
#include <ina219.hpp>

static const char *TAG = "MAIN";

#define uS_TO_S_FACTOR 1000000 /* Conversion factor for micro seconds to seconds */
#define TIME_TO_SLEEP 0 * 1    /* Time ESP32 will go to sleep (in seconds) */

ForteRS485 rs485;
ForteINA219 ina219;
SEM404 rainGaugeSensor = SEM404(2, 19);

float getBatteryVoltage() {
    //************ Measuring Battery Voltage ***********
    // reference voltage of microcontroller 3.3v for esp32
    const float reference_vcc = 3.3;
    // value of R2 resistor [kOhm]
    const float bat_res_gnd = 20;
    // value of R1 resistor [kOhm]
    const float bat_res_vcc = 68;
    // max ADC value
    const int adc = 4095;
    float sample = 0;
    // Get 100 analog read to prevent unusefully read
    for (int i = 0; i < 100; i++) {
        sample = sample + analogRead(35); // read the voltage from the divider circuit
        delay(2);
    }
    sample = sample / 100;
    float battery_voltage = (sample / 4095 * reference_vcc * (bat_res_vcc + bat_res_gnd) / bat_res_gnd);
    ESP_LOGI(TAG, "Battery Voltage: %4.2f V", battery_voltage);
    ESP_LOGD(TAG, "ADC mean Value:  %4.2f", sample);
    return battery_voltage;
}

void setup() {
    // whole loop should be around ~3000 ms
    Serial.begin(115200);
    rs485.setup();

    getBatteryVoltage();

    ina219.setPinSetup({21, 22});
    ina219.setup();

    try {
        ForteRS485::powerOnRS485Sensors();
        auto messages = rainGaugeSensor.buildMessages();
        ForteRS485::powerOffRS485Sensors();
        // split into arrays of 6 messages TODO: Make this cleaner with default constructor
        std::array<Message, 6> messages_1 = {Message::nullMessage(), Message::nullMessage(), Message::nullMessage(),
                                             Message::nullMessage(), Message::nullMessage(), Message::nullMessage()};

        auto ina219_messages = ina219.buildMessages();

        std::array<Message, 6> ina_array = {Message::nullMessage(), Message::nullMessage(), Message::nullMessage(),
                                            Message::nullMessage(), Message::nullMessage(), Message::nullMessage()};

        // log ina messages
        for (const auto &msg : ina219_messages) {
            ESP_LOGI(TAG, "INA219 Message: %s", msg.getMessageAsMinifiedJsonString().c_str());
        }

        int i = 0;
        for (const auto &msg : messages) {
            if (i < 6) {
                messages_1[i] = msg;
            }
            i++;
        }

        espnow_setup();

        // demo of how to reset the rain gauge
#ifdef RESET_RAIN_GAUGE
        if (messages_1.front().getClientDataPackage().getMeasurementData().getValue() > 20) {
            ForteRS485::powerOnRS485Sensors();
            if (rainGaugeSensor.resetPrecipitation() == ErrorType::SEM404_COULD_NOT_RESET_PRECIPITATION) {
                ESP_LOGE(TAG, "Could not reset precipitation");
                auto precipitationErrorMessage = Message(
                    Measurement(ERROR_VALUE, rainGaugeSensor.address, NO_I2C_ADDRESS, MeasurementType::PRECIPITATION,
                                ErrorType::SEM404_COULD_NOT_RESET_PRECIPITATION),
                    rainGaugeSensor.getSensorInformation(), Time::getInstance().getEpochSeconds());
                std::array<Message, 6> messages_error = {precipitationErrorMessage, Message::nullMessage(),
                                                         Message::nullMessage(),    Message::nullMessage(),
                                                         Message::nullMessage(),    Message::nullMessage()};
                Message::sendMessages(messages_error);
            }
            ForteRS485::powerOffRS485Sensors();
        }
#endif

        i = 0;
        for (const auto &msg : ina219_messages) {
            ina_array[i] = msg;
            i++;
        }

        Message::sendMessages(messages_1);
        //    Message::sendMessages(ina_array);

    } catch (const NoDataAvailableException &e) {
        std::cerr << e.what() << '\n';
    }

    rs485.teardown();

    ESP_LOGD(TAG, "Going to sleep for %d seconds", TIME_TO_SLEEP);

    esp_sleep_enable_timer_wakeup(TIME_TO_SLEEP * uS_TO_S_FACTOR);
    esp_deep_sleep_start();
}

void loop() {}
