#include "ESPNow.hpp"
#include "MockSensor.hpp"
#include "NoDataAvailableException.hpp"
#include "f_deep_sleep.hpp"
#include <Arduino.h>
#include <ClientDataPackage.hpp>
#include <soc/rtc_cntl_reg.h>


static const char *TAG = "MAIN";

MockSensor mock_channel0;
MockSensor mock_channel1;
MockSensor mock_channel2;
MockSensor mock_channel3;

void send_msgs(const std::__cxx11::list<Message> msgs) {
    for (const Message &msg: msgs) {

        if (msg.send() != ESP_OK) {
            RtcMemory::getInstance().store_data(msg.getMessageAsMinifiedJsonString());
        }
        unsigned long ts = millis();
        // it takes ~110ms for receiving an acknowledgement by the host in perfect conditions
        uint16_t message_timeout = 2000;
        while (!was_msg_received()) {
            if ((millis() - ts) > message_timeout) {
                RtcMemory::getInstance().store_data(msg.getMessageAsMinifiedJsonString());
                ESP_LOGE(TAG, "Timeout: Host not available\n");
                break;
            }
        }
        ESP_LOGD(TAG, "Time until acknowledgement: %ld", millis() - ts);
    }
}

// one loop takes ~2200 ms
void setup() {
    // disable brownout
//    WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);
    unsigned long ts = millis();
    Serial.begin(115200);

    DeepSleep::print_wakeup_reason();
    DeepSleep::bootCount++;
    ESP_LOGD(TAG, "Boot number: %d", DeepSleep::bootCount);

    // delay(100);
    mock_channel0.setup();
    mock_channel1.setup();
    mock_channel2.setup();
    mock_channel3.setup();

    mock_channel0.setChannel(0);
    mock_channel1.setChannel(1);
    mock_channel2.setChannel(2);
    mock_channel3.setChannel(3);

  // disable led
  gpio_set_direction(GPIO_NUM_32, GPIO_MODE_OUTPUT);

  ESP_LOGD(TAG, "Setup took %ld ms", millis() - ts);

  try {
      // FIXME: put me into seperate trys? No data will be sent when 1 exception occurs
        ts = millis();
        auto messages0 = mock_channel0.buildMessages();
        auto messages1 = mock_channel1.buildMessages();
        auto messages2 = mock_channel2.buildMessages();
        auto messages3 = mock_channel3.buildMessages();

        // roughly takes 500ms, ~120ms for each adc channel, barely anything for battery monitor
        ESP_LOGD(TAG,
                 "Reading data and building messages took %ld ms",
                 millis() - ts);
        gpio_set_level(GPIO_NUM_32, 0);

        ESP_LOGD(TAG, "Size of message to be sent: %d", sizeof(messages0.front()));
        ESP_LOGD(TAG, "Size of Message class: %d", sizeof(Message));
        ESP_LOGD(TAG,
                 "Size of ClientDataPackage class: %d",
                 sizeof(ClientDataPackage));
        // sizeof string
        ESP_LOGD(TAG, "Size of string: %d", sizeof(std::string));
        // sizeof string with 5 char
        ESP_LOGD(TAG, "Size of string with 5 char: %d", sizeof(char[5]));
        // sizeof optional int
        ESP_LOGD(TAG, "Size of optional int: %d", sizeof(std::optional<int>));
        // sizeof int
        ESP_LOGD(TAG, "Size of int: %d", sizeof(int));
        // sizeof short
        ESP_LOGD(TAG, "Size of short: %d", sizeof(short));
        // sizeof double
        ESP_LOGD(TAG, "Size of double: %d", sizeof(double));
        // sizeof float
        ESP_LOGD(TAG, "Size of float: %d", sizeof(float));
        // list of 5 compresseddatapackage
        ClientDataPackage compresseddatapackage{};
        ESP_LOGD(TAG,
                 "Size of list of 4 ClientDataPackage: %d",
                 sizeof(std::list<ClientDataPackage>)
                 + (sizeof(ClientDataPackage) * 4));
        // sizeof compresseddatapackage
        ESP_LOGD(TAG,
                 "Size of ClientDataPackage: %d",
                 sizeof(ClientDataPackage));
        // sizeof list
        ESP_LOGD(TAG, "Size of list: %d", sizeof(std::list<ClientDataPackage>));
        // sizeof vector
        ESP_LOGD(TAG,
                 "Size of vector: %d",
                 sizeof(std::vector<ClientDataPackage>));

        // FIXME: put this outside the try loop?
        ts = millis();
        espnow_setup();
        ESP_LOGD(TAG, "EPSNow setup took %ld ms", millis() - ts);
        // make a list of messages
        std::array<Message, 6> messages =
                {messages0.front(), messages1.front(), messages2.front(),
                 Message::nullMessage(), Message::nullMessage(),
                 Message::nullMessage()};

        // print messages
        for (const Message &msg: messages) {
            ESP_LOGD(TAG, "Message: %s", msg.getMessageAsMinifiedJsonString().c_str());
        }

        ts = millis();
        Message::sendMessages(messages);
        ESP_LOGD(TAG, "Sending messages took %ld ms", millis() - ts);

    } catch (const NoDataAvailableException &e) {
        std::cerr << e.what() << '\n';
    }

    // battery protection: go to deep sleep for unlimited time when voltage less than 3.2V
    DeepSleep::deep_sleep(5);

}

void loop() {}