#include "NoDataAvailableException.hpp"
#include "SentecRainGaugeSensor.h"
#include <SentecSensorsRS485.h>

/***************************************
 *          RS485 SENSOR READOUT
 ****************************************/
static const char *TAG = "SENTEC";

SentecSensorRS485::SentecSensorRS485(byte add) {
    address = add;
}

SentecSensorRS485::SentecSensorRS485(byte add, uint8_t serialControlPin) {
    address = add;
    serialCommunicationControlPin = serialControlPin;
}

void SentecSensorRS485::write(byte queryFrame[], int length) {
    // sends a message (bytes) to the sensor

    // Initialize the transmitter
    digitalWrite(serialCommunicationControlPin, HIGH);
    // Send message: request a reading from the sensor
    RS485->write(queryFrame, length);
    RS485->flush();
    // Initialize the receiver
    digitalWrite(serialCommunicationControlPin, LOW);
}

String SentecSensorRS485::getValueStr(float value) {
    return String(value, 1);
}

String SentecSensorRS485::getValueStr(int value) {
    return String(value);
}

ErrorType SentecSensorRS485::queryAddress() {
    // request the address of the sensor with ONLY ONE SENSOR ON THE BUS

    byte tmp_addr = address; // store the address in a temporary byte
    address = 0xFF;          // change the address to FF (0) for address check
    ErrorType tmp = readRegister(word(0x07, 0xD0), 2);
    address = tmp_addr; // set the original address back
    return tmp;
}

ErrorType SentecSensorRS485::readRegister(int registerStartAddress) {
    return readRegister(registerStartAddress, 1);
}

ErrorType SentecSensorRS485::readRegister(int registerStartAddress, int registerLength) {
    // function code 0x03: get data measured by the sensor
    byte query[8];
    query[0] = address;
    // function code
    query[1] = 0x03;
    // register start address
    query[2] = registerStartAddress >> 8;
    query[3] = registerStartAddress & 0xFF;
    // register length
    query[4] = registerLength >> 8;
    query[5] = registerLength & 0xFF;
    // calculate last two bytes (CRC check)
    calculateCRC(query, sizeof(query) - 2);
    // #   Serial.print("Query (get data): 0x");                    #Print bytes
    // #   printBytes(query, 8);
    //  write the data request to the modbus line
    write(query, sizeof(query));
    // get response from sensor
    return getResponse();
}

WriteRegisterReturnType SentecSensorRS485::writeRegister(int registerAddress, int value) {
    // function code 0x06: change sensor settings
    // e.g. a new address, reset rainfal data...

    byte query[8];
    query[0] = address;
    // function code
    query[1] = 0x06;
    // register start address
    query[2] = registerAddress >> 8;
    query[3] = registerAddress & 0xFF;
    // register length
    query[4] = value >> 8;
    query[5] = value & 0xFF;
    calculateCRC(query, sizeof(query) - 2);
    ESP_LOGD(TAG, "Query (settings):    ");
    printBytes(query, 8);
    write(query, sizeof(query));
    return {getResponse(), {query[0], query[1], query[2], query[3], query[4], query[5], query[6], query[7]}};
}

ErrorType SentecSensorRS485::setAddress(byte add) {
    // change the address of a sensor
    WriteRegisterReturnType tmp = writeRegister(word(0x07, 0xD0), add);
    if (tmp.errorType == ErrorType::DATA_OK)
        address = add;

    if (!isAnswerFrameEqualToQueryFrame(answerFrame, tmp.writtenQuery)) {
        return ErrorType::COULD_NOT_SET_ADDRESS;
    }

    return tmp.errorType;
}

void SentecSensorRS485::resetAnswerFrame() {
    for (unsigned char &i : answerFrame) {
        i = 0;
    }
}

ErrorType SentecSensorRS485::getResponse() {
    // reads the response of a sensor
    ErrorType returnCode = ErrorType::DATA_OK;
    int idx = 0;
    // usual response length: changed in the while loop to match the response,
    //   changed only when reading data (then it's 7 or 9  bytes, sensor dpendent)
    int responseLength = 8;
    // reading an answer takes up to 39 milliseconds for 2 byte readRegister
    const int retries = 1; // #editet to q
    // it doesn't seem to help to request multiple times if first time goes wrong
    for (size_t tries = 1; tries <= retries; tries++) {
        // if we lose connection with the sensor, we get an array of zeros back
        resetAnswerFrame();

        readRS485(idx, responseLength);

        ESP_LOGD(TAG, "Response: 0x%s", formatBytes(answerFrame, responseLength).c_str());

        // TODO: Test this @future_zoe
        returnCode = checkForErrors(idx, responseLength);

        // breaking after first successfull try
        if (returnCode == ErrorType::DATA_OK) {
            break;
        }
    }

    return returnCode;
}

ErrorType SentecSensorRS485::checkForErrors(int idx, int responseLength) {
    // TODO: Not optimal, has precedence
    auto connectionEndedPrematurelyReturnCode = checkConnectionEndedPrematurely(idx, responseLength);
    auto sensorConnectedReturnCode = checkSensorConnected();
    auto crcReturnCode = checkCRC(responseLength);

    if (sensorConnectedReturnCode != ErrorType::DATA_OK) {
        return sensorConnectedReturnCode;
    } else if (connectionEndedPrematurelyReturnCode != ErrorType::DATA_OK) {
        return connectionEndedPrematurelyReturnCode;
    } else if (crcReturnCode != ErrorType::DATA_OK) {
        return crcReturnCode;
    }
    return ErrorType::DATA_OK;
}

ErrorType SentecSensorRS485::checkCRC(int responseLength) {
    word crc_received = word(answerFrame[responseLength - 2], answerFrame[responseLength - 1]);
    word crc = calculateCRC(answerFrame, responseLength - 2);
    if (crc_received != word(crc)) {
        ESP_LOGE(TAG, "CRC wrong: Expected 0x%s got 0x%s", formatBytes(crc).c_str(), formatBytes(crc_received).c_str());
        return ErrorType::WRONG_CRC;
    }
    return ErrorType::DATA_OK;
}

ErrorType SentecSensorRS485::checkConnectionEndedPrematurely(int idx, int responseLength) const {
    if (idx < responseLength) {
        ESP_LOGE(TAG, "Response too short: %d bytes < %d bytes. Unfinished transmission.", idx, responseLength);
        return ErrorType::CONNECTION_ENDED_PREMATURELY;
    }
    return ErrorType::DATA_OK;
}

ErrorType SentecSensorRS485::checkSensorConnected() const {
    if (answerFrame[0] == 0) {
        ESP_LOGE(TAG, "Check sensor connection. First byte is 0x00.");
        return ErrorType::SENSOR_NOT_CONNECTED;
    }
    return ErrorType::DATA_OK;
}

void SentecSensorRS485::readRS485(int &idx, int &responseLength, const int timeout) {
    int byteReceived = 0;
    unsigned long time = millis();
    while (idx < responseLength && (millis() - time) < timeout) {
        if (RS485->available()) {
            byteReceived = RS485->read();
            // Serial.println(byteReceived, HEX);
            // check for first byte. It has to be the device address unless for broadcasts with address = 0xFF
            if (idx == 0 && address != 0xFF && byteReceived != address) {
                ESP_LOGE(TAG, "Invalid byte. First byte needs to be address 0x%02X but got 0x%02Xinstead");
            } else {
                answerFrame[idx] = byteReceived;
                // for reading register: third received byte is data length, read number of bytes accordingly
                if (idx == 2 && answerFrame[1] == 0x03) {
                    // 5 bytes for address, function code, data length, CRC_H, CRC_L
                    responseLength = 5 + byteReceived;
                }
                idx++;
            }
        }
    }
}

unsigned int SentecSensorRS485::calculateCRC(byte query[], int length) {
    // Change the last two bytes of the queryFrame to conform to a CRC check
    // Yes, this is necessary. No, I don't know exactly what it does.
    unsigned int tmp1, tmp2, flag;
    tmp1 = 0xFFFF;
    for (unsigned char i = 0; i < length; i++) {
        tmp1 = tmp1 ^ query[i];
        for (unsigned char j = 1; j <= 8; j++) {
            flag = tmp1 & 0x0001;
            tmp1 >>= 1;
            if (flag)
                tmp1 ^= 0xA001;
        }
    }
    // Reverse byte order.
    tmp2 = tmp1 >> 8;
    tmp1 = (tmp1 << 8) | tmp2;
    tmp1 &= 0xFFFF;
    // change last two query bytes
    query[length + 1] = tmp1;
    query[length] = tmp1 >> 8;

    return tmp1; // the returned value is already swapped - CRC_L byte is first & CRC_H byte is last
}

void SentecSensorRS485::printBytes(byte *data, int length) {
    // prints 8-bit data in hex with leading zeroes
    char tmp[16];
    for (int i = 0; i < length; i++) {
        sprintf(tmp, "%.2X", data[i]);
        // sprintf(tmp, "0x%.2X",data[i]);
        Serial.print(tmp);
        Serial.print(" ");
    }
    Serial.println();
}

void SentecSensorRS485::printBytes(word data) {
    char tmp[10];
    sprintf(tmp, "0x%.2X", data >> 8);
    Serial.print(tmp);
    sprintf(tmp, "%.2X", data & 0xFF);
    // sprintf(tmp, "0x%.2X",data[i]);
    Serial.print(tmp);
}

std::string SentecSensorRS485::formatBytes(byte *data, int length) {
    // pls don't hate me for building strings like that.
    // I also wish that I would be good at programming
    std::string tmp;
    // prints 8-bit data in hex with leading zeroes
    char buff[4];
    for (int i = 0; i < length; i++) {
        snprintf(buff, sizeof(buff), "%02X ", data[i]);
        tmp += buff;
    }
    return tmp;
}

std::string SentecSensorRS485::formatBytes(word data) {
    byte arr[2] = {static_cast<byte>(data >> 8), static_cast<byte>(data & 0xFF)};
    return formatBytes(arr, 2);
}

bool SentecSensorRS485::isAnswerFrameEqualToQueryFrame(byte returnedFrame[8], byte queryFrame[8]) {
    for (int i = 0; i < 8; i++) {
        if (returnedFrame[i] != queryFrame[i]) {
            ESP_LOGD(TAG, "Response: 0x%s", formatBytes(returnedFrame, 8).c_str());
            ESP_LOGD(TAG, "Response: 0x%s", formatBytes(queryFrame, 8).c_str());
            ESP_LOGE("TEST", "Error: returnedFrame[%d] != ret.writtenQuery[%d]", i, i);
            return false;
        }
    }
    return true;
}