#ifndef SENTECSENSORS_H
#define SENTECSENSORS_H

#include <Arduino.h>
#include <ErrorTypes.h>
#include <HardwareNames.h>
#include <RS485HardwareSerial.h>
#include <SensorInformation.hpp>
#include <memory>

struct WriteRegisterReturnType {
    ErrorType errorType;
    byte writtenQuery[8];
};

class SentecSensorRS485 {
  public:
    byte address;
    uint8_t serialCommunicationControlPin = 19;
    std::shared_ptr<HardwareSerial> RS485 = RS485HardwareSerial::getInstance().getRS485Serial();
    byte answerFrame[10] = {0};

    SentecSensorRS485(byte add);

    SentecSensorRS485(byte add, uint8_t serialControlPin);

    void write(byte queryFrame[], int length);

    String getValueStr(float value);

    String getValueStr(int value);

    ErrorType queryAddress();

    ErrorType readRegister(int registerStartAddress);

    ErrorType readRegister(int registerStartAddress, int registerLength);

    WriteRegisterReturnType writeRegister(int registerAddress, int value);

    /**
     * @brief Set the address of the sensor
     * @param add The new address of the sensor
     * @return ErrorType. If the address was not set correctly, the error type is set to
     * ErrorType::COULD_NOT_SET_ADDRESS
     */
    ErrorType setAddress(byte add);

    void resetAnswerFrame();

    ErrorType getResponse();

    unsigned int calculateCRC(byte query[], int length);

    void printBytes(byte *data, int length);

    void printBytes(word data);

    std::string formatBytes(byte *data, int length);

    std::string formatBytes(word data);

  protected:
    /**
     * @brief Check if the answer frame is equal to the query frame. ONLY CHECKS THE FIRST 8 BYTES
     * @param returnedFrame The frame returned from the device
     * @param queryFrame The frame sent to the device
     * @return true if the answer frame is equal to the query frame
     */
    bool isAnswerFrameEqualToQueryFrame(byte *returnedFrame, byte *queryFrame);

  private:
    /**
     * @brief Read the response from the RS485 bus
     * @param[in, out] idx The index of the response
     * @param[in, out] responseLength The length of the response
     * @param timeout The timeout in ms. Default is 2000ms
     */
    void readRS485(int &idx, int &responseLength, const int timeout = 2000);
    ErrorType checkSensorConnected() const;
    ErrorType checkConnectionEndedPrematurely(int idx, int responseLength) const;
    ErrorType checkCRC(int responseLength);
    ErrorType checkForErrors(int idx, int responseLength);
};

#endif
